#include <proto.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "gttextview.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtTextView class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define GTABS(x)		((x) < 0 ? -(x) : (x))

#define GTTF_CURSOR		0x0001
#define GTTF_MARK			0x0002
#define GTTF_BLKCURS	0x0004
#define GTTF_CURSFST	0x0008

#define GTTVA_STRING	(GTDSA_MAX_NR)
#define GTTVA_ATOM_NR	(GTDSA_MAX_NR+1)

static Atom selatoms[GTTVA_ATOM_NR];

/************************************************************************\
*
* Create and init textview.
*
\************************************************************************/
GtTextView *GtMkTextView(char *n,GtManager *p,GtTag *tgs)
{
	GtTextView *obj;
	
	if(!(obj=new GtTextView()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtTextView object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init textview.
*
\************************************************************************/
GtTextView *GtMkTextView(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtTextView *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkTextView(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty textview.
*
\************************************************************************/
GtTextView::GtTextView(void)
{
	GtDInit(&hd);
	topl=NULL;
	top=0;
	left=0;
	curtvp=NULL;
	mrktvp=NULL;
	markstr=NULL;
	tflags=GTTF_BLKCURS;
}

/************************************************************************\
*
* Setup textview.
*
\************************************************************************/
sbool GtTextView::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	XCharStruct xcs;
	sbool bf=TRUE;
	
	if(!GtBdrMan::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_CFILL,GTRT_BOOL,&bf);
	if(bf)
		tflags|=GTTF_BLKCURS;
	else
		tflags&=~GTTF_BLKCURS;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_BlockCursor:
				if(t->d.val)
					tflags|=GTTF_BLKCURS;
				else
					tflags&=~GTTF_BLKCURS;
				break;
		}
	}
	this->PrivateGC(GT_Foreground);
	gtfore->SetV(GT_GraphExpose,TRUE,GT_DONE);
	// Setup atoms for selection.
	memcpy(selatoms,dsp->GetStdAtoms(),GTDSA_MAX_NR*sizeof(Atom));
	selatoms[GTTVA_STRING]=XA_STRING;
	// Get maximum size of default font.
	gtfnt->MaxExtents(&xcs);
	fw=xcs.width;
	foff=xcs.ascent;
	fh=foff+xcs.descent;
	return(TRUE);
}

/************************************************************************\
*
* Set textview attributes.
*
\************************************************************************/
sbool GtTextView::Set(GtTag *tgs)
{
	GtTag *t;
	XCharStruct xcs;
	sbool ch=FALSE,fch=FALSE,vch=FALSE;
	int tp,lt;
	
	if(!GtBdrMan::Set(tgs))
		return(FALSE);
	tp=top;
	lt=left;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_BlockCursor:
				if(t->d.val)
					tflags|=GTTF_BLKCURS;
				else
					tflags&=~GTTF_BLKCURS;
				vch=TRUE;
				break;
			case GT_Top:
				tp=t->d.val;
				ch=TRUE;
				break;
			case GT_Left:
				lt=t->d.val;
				ch=TRUE;
				break;
			case GT_Font:
			case GT_FontList:
				fch=TRUE;
				break;
		}
	}
	if(fch)
	{
		// Check for change of default font and update sizes.
		gtfnt->MaxExtents(&xcs);
		fw=xcs.width;
		foff=xcs.ascent;
		fh=foff+xcs.descent;
		this->UpdateFont();
		this->Refresh();
	}
	// Check for changed position.
	if(ch)
		this->SetPosition(tp,lt);
	if(vch)
		this->Refresh();
	return(TRUE);
}

/************************************************************************\
*
* Get textview attributes.
*
\************************************************************************/
void GtTextView::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtBdrMan::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_BlockCursor:
				if(tflags & GTTF_BLKCURS)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_Top:
				*ptr=top;
				break;
			case GT_Left:
				*ptr=left;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtTextView::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtTextView")==0)
		return(TRUE);
	else
		return(GtBdrMan::IsSubClassOf(cn));
}

/************************************************************************\
*
* Set textview event mask.
*
\************************************************************************/
void GtTextView::SetEMask(ulong em)
{
	GtBdrMan::SetEMask(em|ExposureMask|ButtonPressMask|ButtonReleaseMask|
												Button1MotionMask);
}

/************************************************************************\
*
* Add callback function.
*
\************************************************************************/
void GtTextView::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	if(tag==GT_ValueChange)
		chg.AddCB(cb,ud);
	else
		GtBdrMan::AddCB(tag,cb,ud);
}

/************************************************************************\
*
* Remove callback function.
*
\************************************************************************/
void GtTextView::RemCB(ulong tag,GtCallB cb)
{
	if(tag==GT_ValueChange)
		chg.RemCB(cb);
	else
		GtBdrMan::RemCB(tag,cb);
}

/************************************************************************\
*
* Add a new line at a given position.
*
\************************************************************************/
ulong GtTextView::AddLine(ulong pos)
{
	GtTVLine *tv,*tmp;
	
	// Get memory for a new line.
	if(!(tv=(GtTVLine *)malloc(sizeof(GtTVLine))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for text line!\n");
		return(0);
	}
	// Setup default values for line.
	GtDInit(&tv->phd);
	tv->xsize=fw;
	tv->ysize=fh;
	// Find position to insert line.
	for(tmp=(GtTVLine *)GtDFirst(&hd); tmp && pos;
			tmp=(GtTVLine *)GtDNext(&hd,&tmp->lnk))
	{
		pos--;
	}
	// Add before other line or last in list.
	if(tmp)
		GtDAddBefore(&tmp->lnk,&tv->lnk);
	else
		GtDAddTail(&hd,&tv->lnk);
	// Recalculate maximum sizes.
	this->MaxLineSizes();
	// Return pointer to caller as integer ID.
	return((ulong)tv);
}

/************************************************************************\
*
* Remove line from textview.
*
\************************************************************************/
void GtTextView::RemLine(ulong l)
{
	GtTVLine *tv;
	GtTVPart *tvp;
	
	if(l)
	{
		// Convert integer ID to pointer.
		tv=(GtTVLine *)l;
		// Remove and free all part on a line.
		while((tvp=(GtTVPart *)GtDFirst(&tv->phd)))
			this->RemPart((ulong)tvp);
		// Remove from list and free.
		GtDRemove(&tv->lnk);
		free(tv);
		// Recalculate maximum sizes.
		this->MaxLineSizes();
	}
}

/************************************************************************\
*
* Remove all lines in textview.
*
\************************************************************************/
void GtTextView::RemAllLines(void)
{
	GtTVLine *tvl;
	
	this->SetPosition(0,0);
	while((tvl=(GtTVLine *)GtDFirst(&hd)))
		this->RemLine((ulong)tvl);
}

/************************************************************************\
*
* Add a string part to a line at a given position.
*
\************************************************************************/
ulong GtTextView::AddPart(ulong l,char *str,int len,ulong pos)
{
	GtTVLine *tvl;
	GtTVPart *tvp,*tmp;
	XCharStruct xcs;
	
	// Get memory for line part.
	if(!(tvp=(GtTVPart *)malloc(sizeof(GtTVPart))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for string part!\n");
		return(0);
	}
	// Limit string length to 256 chars.
	len=GTMIN(len,GTTV_MAXSTRLEN);
	// Set type and defaults of string part.
	tvp->type=GTTV_STRING;
	tvp->d.txt.fnt=NULL;
	tvp->d.txt.attr=0;
	gtfnt->MaxExtents(&xcs);
	tvp->d.txt.yoff=xcs.ascent;
	tvp->ysize=xcs.ascent+xcs.descent;
	tvp->xsize=gtfnt->StrWidth(str,len);
	tvp->d.txt.str=str;
	tvp->d.txt.len=len;
	tvl=(GtTVLine *)l;
	tvp->tvl=tvl;
	// Find position of part within line.
	for(tmp=(GtTVPart *)GtDFirst(&tvl->phd); tmp && pos;
			tmp=(GtTVPart *)GtDNext(&tvl->phd,&tmp->lnk))
	{
		pos--;
	}
	// Add part before other part or last in line.
	if(tmp)
		GtDAddBefore(&tmp->lnk,&tvp->lnk);
	else
		GtDAddTail(&tvl->phd,&tvp->lnk);
	// Recalculate sizes for lines.
	this->MaxPartSizes(tvl);
	this->MaxLineSizes();
	// Return pointer to part as integer ID.
	return((ulong)tvp);
}
	
/************************************************************************\
*
* Add a widget part to a line at a given part.
*
\************************************************************************/
ulong GtTextView::AddPart(ulong l,GtWidget *w,ubyte attr,ulong pos)
{
	GtTVLine *tvl;
	GtTVPart *tvp,*tmp;
	GtWListTV *li=NULL;
	int xs,ys,dy=0;
	
	// Find widget to create part from in list of managed childrean.
	for(li=(GtWListTV *)GtSFirst(&man); li;
			li=(GtWListTV *)GtSNext(&li->bwl.wl.lnk))
	{
		if(li->bwl.wl.w==w)
			break;
	}
	if(!li)
	{
		GtError(this,GTEC_MISC,"Unconnected widget '%s' to add as part!\n",
						w->GetName());
		return(0);
	}
	// Get memory for widget part.
	if(!(tvp=(GtTVPart *)malloc(sizeof(GtTVPart))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for string part!\n");
		return(0);
	}
	// Set type and defaults for widget part.
	li->tvp=tvp;
	tvp->type=GTTV_WIDGET;
	w->GetMinSize(&xs,&ys);
	tvp->xsize=GTMAX(xs,li->bwl.minxs)+li->bwl.bleft+li->bwl.bright;
	tvp->ysize=GTMAX(ys,li->bwl.minys)+li->bwl.btop+li->bwl.bbot;
	tvp->d.w.wl=li;
	tvp->d.w.attr=attr;
	tvl=(GtTVLine *)l;
	tvp->tvl=tvl;
	// Find position of part within line.
	for(tmp=(GtTVPart *)GtDFirst(&tvl->phd); tmp && pos;
			tmp=(GtTVPart *)GtDNext(&tvl->phd,&tmp->lnk))
	{
		pos--;
	}
	// Add part before other part or last in line.
	if(tmp)
		GtDAddBefore(&tmp->lnk,&tvp->lnk);
	else
		GtDAddTail(&tvl->phd,&tvp->lnk);
	// Recalculate sizes for lines.
	this->MaxPartSizes(tvl);
	this->MaxLineSizes();
	// Position widget correctly.
	if((tvl->ypos+tvl->ysize<top) || (tvl->ypos>=top+ysize) ||
			(tvp->xpos+tvp->xsize<left) || (tvp->xpos>=left+xsize))
	{
		// Hide widget by moving it off visible area.
		xs=tvp->xsize-li->bwl.bleft-li->bwl.bright;
		ys=tvp->ysize-li->bwl.btop-li->bwl.bbot;
		li->bwl.wl.w->SetPosSize(-xs,-ys,xs,ys);
	}
	else
	{
		// Position widget in visible area.
		dy=(tvl->ysize-tvp->ysize)/2;
		xs=tvp->xsize-li->bwl.bleft-li->bwl.bright;
		ys=tvp->ysize-li->bwl.btop-li->bwl.bbot;
		li->bwl.wl.w->SetPosSize(
			tvp->xpos-left+li->bwl.bleft,tvl->ypos-top+dy+li->bwl.btop,xs,ys);
	}
	// Return pointer to part as integer ID.
	return((ulong)tvp);
}

/************************************************************************\
*
* Remove part from line.
*
\************************************************************************/
void GtTextView::RemPart(ulong p)
{
	GtTVLine *tvl;
	GtTVPart *tvp;
	GtWListTV *li;
	int xs,ys;
	
	// Convert integer ID to pointer and find the line of the part.
	tvp=(GtTVPart *)p;
	tvl=tvp->tvl;
	if(tvl && tvp)
	{
		// Remove part and free it.
		GtDRemove(&tvp->lnk);
		if((tvp->type==GTTV_WIDGET) && tvp->d.w.wl)
		{
			li=tvp->d.w.wl;
			xs=tvp->xsize-li->bwl.bleft-li->bwl.bright;
			ys=tvp->ysize-li->bwl.btop-li->bwl.bbot;
			li->bwl.wl.w->SetPosSize(-xs,-ys,xs,ys);
			li->tvp=NULL;
			if(tvp->d.w.attr & GTTV_FREE)
				GtDestroy(li->bwl.wl.w);
		}
		if((tvp->type==GTTV_STRING) && (tvp->d.txt.attr & GTTV_FREE))
			free(tvp->d.txt.str);
		free(tvp);
		// Recalculate sizes of lines.
		this->MaxPartSizes(tvl);
		this->MaxLineSizes();
	}
}

/************************************************************************\
*
* Set string part attibutes.
*
\************************************************************************/
void GtTextView::SetTextAttr(ulong p,GtFont *f,ubyte attr,ulong fg,ulong bg)
{
	GtTVPart *tvp;
	XCharStruct xcs;
	
	tvp=(GtTVPart *)p;
	if(tvp && (tvp->type==GTTV_STRING))
	{
		// Set font for part.
		tvp->d.txt.fnt=f;
		if(f)
		{
			// Calculate new sizes for part.
			f->MaxExtents(&xcs);
			tvp->d.txt.yoff=xcs.ascent;
			tvp->ysize=xcs.ascent+xcs.descent;
			tvp->xsize=f->StrWidth(tvp->d.txt.str,tvp->d.txt.len);
			this->MaxPartSizes(tvp->tvl);
			this->MaxLineSizes();
		}
		tvp->d.txt.attr=attr;
		tvp->d.txt.fg=fg;
		tvp->d.txt.bg=bg;
	}
}

/************************************************************************\
*
* Position visible part of text.
*
\************************************************************************/
void GtTextView::SetPosition(int xp,int yp)
{
	GtTVLine *tv;
	GtTVPart *tvp;
	GtWListTV *li;
	int ol,ot,dx,dy;
	word xs,ys;
	
	// Save old position.
	ol=left;
	ot=top;
	// Calculate new position.
	left=GTMAX(GTMIN(xp,lmxs-xsize),0);
	top=GTMAX(GTMIN(yp,lmys-ysize),0);
	// Hide any widget not in visible area.
	for(li=(GtWListTV *)GtSFirst(&man); li;
			li=(GtWListTV *)GtSNext(&li->bwl.wl.lnk))
	{
		// Check if widget is connected to any part.
		if(li->tvp)
		{
			tvp=li->tvp;
			tv=tvp->tvl;
			// Check is part is outside visible area.
			if((tv->ypos+tv->ysize<top) || (tv->ypos>=top+ysize) ||
					(tvp->xpos+tvp->xsize<left) || (tvp->xpos>=left+xsize))
			{
				// Hide widget by moving it of visible area.
				xs=tvp->xsize-li->bwl.bleft-li->bwl.bright;
				ys=tvp->ysize-li->bwl.btop-li->bwl.bbot;
				li->bwl.wl.w->SetPosSize(-xs,-ys,xs,ys);
			}
			else
			{
				// Position widget in visible area.
				dy=(tv->ysize-tvp->ysize)/2;
				xs=tvp->xsize-li->bwl.bleft-li->bwl.bright;
				ys=tvp->ysize-li->bwl.btop-li->bwl.bbot;
				li->bwl.wl.w->SetPosSize(
					tvp->xpos-left+li->bwl.bleft,tv->ypos-top+dy+li->bwl.btop,xs,ys);
			}
		}
	}
	// Any change in position?
	if((ol!=left) || (ot!=top))
	{
		// Calculate amount of change.
		dx=ol-left;
		dy=ot-top;
		if(dy<0)
		{
			// Find new top line.
			for(tv=topl; tv; tv=(GtTVLine *)GtDNext(&hd,&tv->lnk))
			{
				if(tv->ypos+tv->ysize>=top)
				{
					topl=tv;
					break;
				}
				topnr++;
			}
		}
		else
		{
			// Find new top line.
			for(tv=topl; tv; tv=(GtTVLine *)GtDPrev(&hd,&tv->lnk))
			{
				if(tv->ypos<=top)
				{
					topl=tv;
					break;
				}
				topnr--;
			}
		}
		// Check if changes where larger than widget.
		if((GTABS(dx)>=xsize) || (GTABS(dy)>=ysize))
			this->Refresh();
		else
		{
			// Only small change so let's be smart.
			// Scroll text by copy.
			gtfore->CopyArea(gtwin,gtwin,0,0,xsize,ysize,dx,dy);
			// Clear any new area.
			if(dx<0)
			{
				gtwin->Clear(xsize+dx,0,-dx,ysize);
				this->ReDraw(left+xsize+dx,top,-dx,ysize);
			}
			else if(dx>0)
			{
				gtwin->Clear(0,0,dx,ysize);
				this->ReDraw(left,top,dx,ysize);
			}
			if(dy<0)
			{
				gtwin->Clear(0,ysize+dy,xsize,-dy);
				this->ReDraw(left,top+ysize+dy,xsize,-dy);
			}
			else if(dy>0)
			{
				gtwin->Clear(0,0,xsize,dy);
				this->ReDraw(left,top,xsize,dy);
			}
		}
	}
}

/************************************************************************\
*
* Set top position by line id.
*
\************************************************************************/
void GtTextView::SetTopLine(ulong l)
{
	GtTVLine *tvl;
	
	tvl=(GtTVLine *)l;
	if(tvl)
		this->SetPosition(left,tvl->ypos);
}

/************************************************************************\
*
* Set top position by line number.
*
\************************************************************************/
void GtTextView::SetTopLineNr(int nr)
{
	GtTVLine *tvl;
	
	for(tvl=(GtTVLine *)GtDFirst(&hd); tvl && (nr > 0);
			tvl=(GtTVLine *)GtDNext(&hd,&tvl->lnk))
	{
		nr--;
	}
	if(tvl)
		this->SetPosition(left,tvl->ypos);
}

/************************************************************************\
*
* Remove child and part if connected.
*
\************************************************************************/
void GtTextView::RemChild(GtWidget *w)
{
	GtWListTV *li;
	
	li=(GtWListTV *)w->GetLInfo();
	if(li && li->tvp)
	{
		li->tvp->d.w.wl=NULL;
		this->RemPart((ulong)li->tvp);
		this->Refresh();
	}
	GtBdrMan::RemChild(w);
}

/************************************************************************\
*
* Request from child to move or resize.
*
\************************************************************************/
void GtTextView::MoveSizeReq(GtWidget *o,word xp,word yp,word xs,word ys)
{
	GtTVPart *tvp;
	GtWListTV *li;
	int mxs,mys;
	
	li=(GtWListTV *)o->GetLInfo();
	if(li->tvp)
	{
		// Calculate new size of widget.
		tvp=li->tvp;
		li->bwl.wl.w->GetMinSize(&mxs,&mys);
		tvp->xsize=GTMAX(mxs,li->bwl.minxs)+li->bwl.bleft+li->bwl.bright;
		tvp->ysize=GTMAX(mys,li->bwl.minys)+li->bwl.btop+li->bwl.bbot;
		// Update sizes of parts and lines.
		this->MaxPartSizes(tvp->tvl);
		this->MaxLineSizes();
		this->Refresh();
	}
	else
		o->SetPosSize(-xs,-ys,xs,ys);
}

/************************************************************************\
*
* Set textview position and size.
*
\************************************************************************/
void GtTextView::SetPosSize(word xp,word yp,word xs,word ys)
{
	GtTextViewCD cd;
	word oxs,oys;
	int nt,nl;
	
	// Save old size.
	oxs=xsize;
	oys=ysize;
	GtBdrMan::SetPosSize(xp,yp,xs,ys);
	nl=left;
	nt=top;
	// Check if reposition is needed to fill visible area.
	if(lmys-top<ysize)
		nt=GTMAX(0,lmys-ysize);
	if(lmxs-left<xsize)
		nl=GTMAX(0,lmxs-xsize);
	this->SetPosition(nl,nt);
	// Any change in size?
	if((oxs!=xsize) || (oys!=ysize))
	{
		// Call value change callbacks.
		cd.txtxp=left;
		cd.txtyp=top;
		cd.txtxs=lmxs;
		cd.txtys=lmys;
		cd.xsize=xsize;
		cd.ysize=ysize;
		cd.topline=topnr;
		cd.totline=totnr;
		chg.CallCB(this,&cd);
	}
}

/************************************************************************\
*
* Draw textview.
*
\************************************************************************/
sbool GtTextView::Draw(XEvent *xe)
{
	if(xe)
	{
		if(xe->type==Expose)
		{
			this->ReDraw(left+xe->xexpose.x,top+xe->xexpose.y,
										xe->xexpose.width,xe->xexpose.height);
		}
		else if(xe->type==GraphicsExpose)
		{
			this->ReDraw(left+xe->xgraphicsexpose.x,top+xe->xgraphicsexpose.y,
										xe->xgraphicsexpose.width,xe->xgraphicsexpose.height);
		}
	}
	else
		this->ReDraw(left,top,xsize,ysize);
	return(TRUE);
}

/************************************************************************\
*
* Handle mouse events for textview.
*
\************************************************************************/
sbool GtTextView::Mouse(XEvent *xe)
{
	GtTVLine *tvl1,*tvl2;
	
	if(!xe)
		return(TRUE);
	switch(xe->type)
	{
		case ButtonPress:
			// Position cursor position.
			if(xe->xbutton.button==Button1)
			{
				// Unmark previous mark.
				if(tflags & GTTF_MARK)
				{
					this->UpdateMark(FALSE);
					this->Refresh();
					tflags&=~GTTF_MARK;
				}
				curtvp=NULL;
				// Find new cursor position.
				this->FindCursPos(xe->xbutton.x,xe->xbutton.y);
				// Copy cursor position to mark position.
				mrktvp=curtvp;
				mrkpos=curpos;
				mrkxp=curxp;
				mrkxs=curxs;
				this->Refresh();
			}
			break;
		case ButtonRelease:
			if((xe->xbutton.button==Button1) && (tflags & GTTF_MARK))
			{
				// Own selection.
				if(dsp->OwnSelection(this,XA_PRIMARY,xe->xbutton.time,
						ConvMark,LoseMark,DoneMark,this))
				{
					this->AllocMark();
				}
				else
				{
					tflags&=~GTTF_MARK;
					this->UpdateMark(FALSE);
					this->Refresh();
				}
			}
			break;
		case MotionNotify:
			if(!mrktvp)
				break;
			// Unmark previous mark.
			if(tflags & GTTF_MARK)
				this->UpdateMark(FALSE);
			// Find new cursor position.
			tvl1=curtvp->tvl;
			this->FindCursPos(xe->xbutton.x,xe->xbutton.y);
			tvl2=curtvp->tvl;
			// Update new mark.
			if((tflags & GTTF_MARK) || (curtvp!=mrktvp) || (curxp!=mrkxp))
			{
				this->UpdateMark(TRUE);
				// Refresh only changed lines.
				if(tvl1->ypos<=tvl2->ypos)
				{
					this->RefreshRect(0,tvl1->ypos,xsize,
						tvl2->ypos-tvl1->ypos+tvl2->ysize);
				}
				else
				{
					this->RefreshRect(0,tvl2->ypos,xsize,
						tvl1->ypos-tvl2->ypos+tvl1->ysize);
				}
				tflags|=GTTF_MARK;
			}
			break;
	}
	return(TRUE);
}

/************************************************************************\
*
* Create child layout information.
*
\************************************************************************/
GtWList *GtTextView::MakeWList(GtWidget *o,int sz)
{
	GtWListTV *li;
	
	li=(GtWListTV *)GtBdrMan::MakeWList(o,
		sz+sizeof(GtWListTV)-sizeof(GtWListBdr));
	if(!li)
		return(NULL);
	li->tvp=NULL;
	return((GtWList *)li);
}

/************************************************************************\
*
* Find cursor position from window position.
*
\************************************************************************/
void GtTextView::FindCursPos(int xp,int yp)
{
	GtFont *f;
	GtTVLine *tvl;
	GtTVPart *tvp;
	int ixp,iyp;
	
	ixp=xp+left;
	iyp=yp+top;
	// Loop through visible lines.
	for(tvl=topl; tvl && (tvl->ypos<top+ysize);
			tvl=(GtTVLine *)GtDNext(&hd,&tvl->lnk))
	{
		// Check if point is in line.
		if((iyp>=tvl->ypos) && (iyp<tvl->ypos+tvl->ysize))
		{
			// Loop through visible parts in line.
			for(tvp=(GtTVPart *)GtDFirst(&tvl->phd); tvp && (tvp->xpos<left+xsize);
					tvp=(GtTVPart *)GtDNext(&tvl->phd,&tvp->lnk))
			{
				// Check if point is in part.
				if((ixp>=tvp->xpos) && (ixp<tvp->xpos+tvp->xsize) &&
						(tvp->type==GTTV_STRING))
				{
					if(!(f=tvp->d.txt.fnt))
						f=gtfnt;
					curtvp=tvp;
					curpos=0;
					curxp=0;
					// Find char position in part.
					while(tvp->d.txt.str[curpos])
					{
						curxs=f->CharWidth(tvp->d.txt.str[curpos]);
						if((ixp>=tvp->xpos+curxp) && (ixp<tvp->xpos+curxp+curxs))
							return;
						curpos++;
						curxp+=curxs;
					}
				}
			}
			// Point not in any part so set position after last part.
			for(tvp=(GtTVPart *)GtDLast(&tvl->phd); tvp;
					tvp=(GtTVPart *)GtDPrev(&tvl->phd,&tvp->lnk))
			{
				if(tvp->type==GTTV_STRING)
				{
					if(!(f=tvp->d.txt.fnt))
						f=gtfnt;
					curtvp=tvp;
					curpos=tvp->d.txt.len;
					curxp=tvp->xsize;
					curxs=f->CharWidth(' ');
					return;
				}
			}
		}
	}
}

/************************************************************************\
*
* Mark or unmark all parts between mark position and cursor position.
*
\************************************************************************/
void GtTextView::UpdateMark(sbool on)
{
	GtTVLine *tvl1,*tvl2;
	GtTVPart *tvp,*end=NULL;
	sbool count=FALSE;

	marklen=0;
	// Select first and last line.
	tvl1=curtvp->tvl;
	tvl2=mrktvp->tvl;
	if(tvl2->ypos < tvl1->ypos)
	{
		tvl1=mrktvp->tvl;
		tvl2=curtvp->tvl;
	}
	tvl2=(GtTVLine *)GtDNext(&hd,&tvl2->lnk);
	// Loop through all lines marked.
	for(; tvl1 && (tvl1 != tvl2); tvl1=(GtTVLine *)GtDNext(&hd,&tvl1->lnk))
	{
		marklen++;
		// Loop through all parts on marked lines.
		for(tvp=(GtTVPart *)GtDFirst(&tvl1->phd); tvp;
				tvp=(GtTVPart *)GtDNext(&tvl1->phd,&tvp->lnk))
		{
			// Count is false if first mark position not found.
			if(!count)
			{
				// Check if cursor position is found first.
				if(tvp==curtvp)
				{
					count=TRUE;
					end=mrktvp;
					tflags|=GTTF_CURSFST;
				}
				// Check if mark position is found first.
				if(tvp==mrktvp)
				{
					count=TRUE;
					end=curtvp;
					tflags&=~GTTF_CURSFST;
				}
			}
			if(count && (tvp->type==GTTV_STRING))
			{
				// Mark part.
				tvp->d.txt.attr=(on ? tvp->d.txt.attr|GTTV_MARK :
															tvp->d.txt.attr&~GTTV_MARK);
				// Sum length of all marked parts.
				if(curtvp==mrktvp)
					marklen+=GTMAX(curpos-mrkpos,mrkpos-curpos);
				else if(tvp==curtvp)
				{
					if(tflags & GTTF_CURSFST)
						marklen+=tvp->d.txt.len-curpos;
					else
						marklen+=curpos+1;
				}
				else if(tvp==mrktvp)
				{
					if(tflags & GTTF_CURSFST)
						marklen+=mrkpos+1;
					else
						marklen+=tvp->d.txt.len-mrkpos;
				}
				else
					marklen+=tvp->d.txt.len;
			}
			// Stop when last marked part found.
			if(tvp==end)
				break;
		}
	}
}

/************************************************************************\
*
* Refresh rectangle of textview.
*
\************************************************************************/
void GtTextView::RefreshRect(int xp,int yp,int xs,int ys)
{
	if(gfx)
	{
		gtwin->Clear(xp-left,yp-top,xs,ys);
		this->ReDraw(xp,yp,xs,ys);
	}
}

/************************************************************************\
*
* Redraw a visible area of the text.
*
\************************************************************************/
void GtTextView::ReDraw(int xp,int yp,int xs,int ys)
{
	GtFont *f,*lf;
	GtTVLine *tvl;
	GtTVPart *tvp;
	ulong fg,bg,lfg,lbg,tmp;
	ubyte attr;
	
	// Check to make sure some refreshing will occure.
	if((xp>=left+xsize) || (yp>=top+ysize) || (xp+xs<left) || (yp+ys<top))
		return;
	lf=gtfnt;
	lfg=gfx->cols[GTFORE];
	lbg=gfx->cols[GTBACK];
	// Start at top line and work down to last line in refresh area.
	for(tvl=topl; tvl && (tvl->ypos<yp+ys);
			tvl=(GtTVLine *)GtDNext(&hd,&tvl->lnk))
	{
		// Is line in refresh area?
		if(tvl->ypos+tvl->ysize>yp)
		{
			// Start at first part and work right to last part in refresh area.
			for(tvp=(GtTVPart *)GtDFirst(&tvl->phd); tvp && (tvp->xpos<xp+xs);
					tvp=(GtTVPart *)GtDNext(&tvl->phd,&tvp->lnk))
			{
				// Is part in refresh area?
				if(tvp->xpos+tvp->xsize>xp)
				{
					// Type of part?
					if(tvp->type==GTTV_STRING)
					{
						attr=tvp->d.txt.attr;
						// Check for special font.
						if(!(f=tvp->d.txt.fnt))
							f=gtfnt;
						// Change font if new one.
						if(f!=lf)
						{
							gtfore->SetFont(f);
							lf=f;
						}
						// Select foreground and background colors.
						if(attr & GTTV_FG)
							fg=tvp->d.txt.fg;
						else
							fg=gfx->cols[GTFORE];
						if(attr & GTTV_BG)
							bg=tvp->d.txt.bg;
						else
							bg=gfx->cols[GTBACK];
						// Change colors if new ones.
						if(lfg!=fg)
						{
							gtfore->SetForeground(fg);
							lfg=fg;
						}
						if(lbg!=bg)
						{
							gtfore->SetBackground(bg);
							lbg=bg;
						}
						// Check if this part is marked.
						if(attr & GTTV_MARK)
						{
							if(curtvp==mrktvp)
							{
								gtlht->DrawRectangle(gtwin,tvp->xpos-left+GTMIN(curxp,mrkxp),
									tvl->ypos-top,GTMAX(curxp-mrkxp,mrkxp-curxp)+
									(tflags & GTTF_CURSFST ? mrkxs : curxs),tvl->ysize,TRUE);
							}
							else if(curtvp==tvp)
							{
								gtlht->DrawRectangle(gtwin,
									tvp->xpos-left+(tflags & GTTF_CURSFST ? curxp : 0),tvl->ypos-top,
									(tflags & GTTF_CURSFST ? tvp->xsize-curxp : curxp+curxs),
									tvl->ysize,TRUE);
							}
							else if(mrktvp==tvp)
							{
								gtlht->DrawRectangle(gtwin,
									tvp->xpos-left+(tflags & GTTF_CURSFST ? 0: mrkxp),tvl->ypos-top,
									(tflags & GTTF_CURSFST ? mrkxp+mrkxs : tvp->xsize-mrkxp),
									tvl->ysize,TRUE);
							}
							else
							{
								gtlht->DrawRectangle(gtwin,tvp->xpos-left,tvl->ypos-top,
									tvp->xsize,tvl->ysize,TRUE);
							}
						}
						// Reverse colors and set filling.
						if(attr & GTTV_REVERSE)
						{
							gtfore->DrawRectangle(gtwin,tvp->xpos-left,tvl->ypos-top,
								tvp->xsize,tvl->ysize,TRUE);
							tmp=fg;
							fg=bg;
							bg=tmp;
							gtfore->SetForeground(fg);
							gtfore->SetBackground(bg);
							lfg=fg;
							lbg=bg;
						}
						// Check if cursor is in this part
						if((tflags & GTTF_CURSOR) && (tvp == curtvp))
						{
							if(tflags & GTTF_BLKCURS)
							{
								gtfill->DrawRectangle(gtwin,tvp->xpos+curxp-left,tvl->ypos-top,
									curxs,tvl->ysize,TRUE);
							}
							else
							{
								gtfill->DrawLine(gtwin,tvp->xpos+curxp-left,tvl->ypos-top,
									tvp->xpos+curxp-left,tvl->ypos-top+tvl->ysize-1);
							}
						}
						// Draw string.
						gtfore->DrawString(gtwin,tvp->xpos-left,tvl->ypos-top+tvl->yoff,
							tvp->d.txt.str,tvp->d.txt.len,FALSE);
						if(attr & GTTV_BOLD)
						{
							// Shift right on pixel and draw again to create bold.
							gtfore->DrawString(gtwin,1+tvp->xpos-left,tvl->ypos-top+tvl->yoff,
								tvp->d.txt.str,tvp->d.txt.len,FALSE);
						}
						if(attr & GTTV_ULINE)
						{
							// Draw line below string for underline.
							gtfore->DrawLine(gtwin,tvp->xpos-left,tvl->ypos-top+tvl->yoff+1,
								tvp->xpos-left+tvp->xsize-1,tvl->ypos-top+tvl->yoff+1);
						}
					}
				}
			}
		}
	}
	// Restore default values of GC.
	if(lf!=gtfnt)
		gtfore->SetFont(gtfnt);
	if(lfg!=gfx->cols[GTFORE])
		gtfore->SetForeground(gfx->cols[GTFORE]);
	if(lbg!=gfx->cols[GTBACK])
		gtfore->SetForeground(gfx->cols[GTBACK]);	
}

/************************************************************************\
*
* Calculate new maximum sizes and set line positions.
*
\************************************************************************/
void GtTextView::MaxLineSizes(void)
{
	GtTVLine *tv;
	GtTextViewCD cd;
	int nt,nl;
	
	lmxs=1;
	lmys=0;
	totnr=0;
	topnr=0;
	topl=NULL;
	// Calculate maximum size of text.
	for(tv=(GtTVLine *)GtDFirst(&hd); tv;
			tv=(GtTVLine *)GtDNext(&hd,&tv->lnk))
	{
		tv->ypos=lmys;
		lmxs=GTMAX(lmxs,tv->xsize);
		lmys+=tv->ysize;
		// Set top line if not set before.
		if(!topl && (tv->ypos+tv->ysize>=top))
			topl=tv;
		if(!topl)
			topnr++;
		totnr++;
	}
	nl=left;
	nt=top;
	// Check if reposition is needed to fill visible area.
	if(lmys-top<ysize)
		nt=GTMAX(0,lmys-ysize);
	if(lmxs-left<xsize)
		nl=GTMAX(0,lmxs-xsize);
	this->SetPosition(nl,nt);
	cd.txtxp=left;
	cd.txtyp=top;
	cd.txtxs=lmxs;
	cd.txtys=lmys;
	cd.xsize=xsize;
	cd.ysize=ysize;
	cd.topline=topnr;
	cd.totline=totnr;
	chg.CallCB(this,&cd);
}

/************************************************************************\
*
* Calculate new maximum sizes of line parts.
*
\************************************************************************/
void GtTextView::MaxPartSizes(GtTVLine *tvl)
{
	GtTVPart *tvp;
	int asc,desc;
	
	tvl->xsize=0;
	tvl->ysize=fh;
	asc=foff;
	desc=fh-foff;
	// Calculate size of line by adding maximum ascent and descent.
	for(tvp=(GtTVPart *)GtDFirst(&tvl->phd); tvp;
			tvp=(GtTVPart *)GtDNext(&tvl->phd,&tvp->lnk))
	{
		tvp->xpos=tvl->xsize;
		tvl->xsize+=tvp->xsize;
		if(tvp->type==GTTV_STRING)
		{
			asc=GTMAX(asc,tvp->d.txt.yoff);
			desc=GTMAX(desc,tvp->ysize-tvp->d.txt.yoff);
		}
		else
			tvl->ysize=GTMAX(tvl->ysize,tvp->ysize);
	}
	if((tflags & GTTF_CURSOR) && (tvp=(GtTVPart *)GtDLast(&tvl->phd)) &&
			(tvp->type==GTTV_STRING))
	{
		if(tvp->d.txt.fnt)
			tvl->xsize+=tvp->d.txt.fnt->CharWidth(' ');
		else
			tvl->xsize+=gtfnt->CharWidth(' ');
	}
	tvl->yoff=asc;
	tvl->ysize=GTMAX(tvl->ysize,asc+desc);
}

/************************************************************************\
*
* Update default font.
*
\************************************************************************/
void GtTextView::UpdateFont(void)
{
	GtTVLine *tvl;
	GtTVPart *tvp;
	XCharStruct xcs;
	sbool ch,lch=FALSE;
	
	for(tvl=(GtTVLine *)GtDFirst(&hd); tvl;
			tvl=(GtTVLine *)GtDNext(&hd,&tvl->lnk))
	{
		ch=FALSE;
		for(tvp=(GtTVPart *)GtDFirst(&tvl->phd); tvp;
				tvp=(GtTVPart *)GtDNext(&tvl->phd,&tvp->lnk))
		{
			if((tvp->type==GTTV_STRING) && !tvp->d.txt.fnt)
			{
				gtfnt->MaxExtents(&xcs);
				tvp->d.txt.yoff=xcs.ascent;
				tvp->ysize=xcs.ascent+xcs.descent;
				tvp->xsize=gtfnt->StrWidth(tvp->d.txt.str,tvp->d.txt.len);
				ch=TRUE;
			}
		}
		if(ch)
		{
			this->MaxPartSizes(tvl);
			lch=TRUE;
		}
	}
	if(lch)
		this->MaxLineSizes();
}

/************************************************************************\
*
* Allocate and copy marked text.
*
\************************************************************************/
void GtTextView::AllocMark(void)
{
	GtTVLine *tvl1,*tvl2;
	GtTVPart *tvp,*end=NULL;
	int len=0,pos=0;
	sbool count;
	
	// Allocate memory for marked text.
	if(!(markstr=(char *)malloc(marklen+1)))
		return;
	count=FALSE;
	tvl1=curtvp->tvl;
	tvl2=mrktvp->tvl;
	if(tvl2->ypos < tvl1->ypos)
	{
		tvl1=mrktvp->tvl;
		tvl2=curtvp->tvl;
	}
	tvl2=(GtTVLine *)GtDNext(&hd,&tvl2->lnk);
	// Loop through all marked lines.
	for(; tvl1 && (tvl1 != tvl2); tvl1=(GtTVLine *)GtDNext(&hd,&tvl1->lnk))
	{
		for(tvp=(GtTVPart *)GtDFirst(&tvl1->phd); tvp;
				tvp=(GtTVPart *)GtDNext(&tvl1->phd,&tvp->lnk))
		{
			if(!count)
			{
				if(tvp==curtvp)
				{
					count=TRUE;
					end=mrktvp;
					tflags|=GTTF_CURSFST;
				}
				if(tvp==mrktvp)
				{
					count=TRUE;
					end=curtvp;
					tflags&=~GTTF_CURSFST;
				}
			}
			if(count && (tvp->type==GTTV_STRING))
			{
				// Copy all marked parts to allocated memory.
				if(curtvp==mrktvp)
				{
					len=GTMAX(curpos-mrkpos,mrkpos-curpos);
					strncpy(markstr+pos,tvp->d.txt.str+GTMIN(curpos,mrkpos),len);
				}
				else if(tvp==curtvp)
				{
					if(tflags & GTTF_CURSFST)
					{
						len=tvp->d.txt.len-curpos;
						strncpy(markstr+pos,tvp->d.txt.str+curpos,len);
					}
					else
					{
						len=curpos+1;
						strncpy(markstr+pos,tvp->d.txt.str,len);
					}
				}
				else if(tvp==mrktvp)
				{
					if(tflags & GTTF_CURSFST)
					{
						len=mrkpos+1;
						strncpy(markstr+pos,tvp->d.txt.str,len);
					}
					else
					{
						len=tvp->d.txt.len-mrkpos;
						strncpy(markstr+pos,tvp->d.txt.str+mrkpos,len);
					}
				}
				else
				{
					len=tvp->d.txt.len;
					strncpy(markstr+pos,tvp->d.txt.str,len);
				}
				pos+=len;
				if(pos && (markstr[pos-1]=='\0'))
					pos--;
			}
			if(tvp==end)
				break;
		}
		markstr[pos]='\n';
		pos++;
	}
	markstr[pos]='\0';
}

/************************************************************************\
*
* Convert selection request.
*
\************************************************************************/
int GtTextView::ConvMark(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtTextView *tv;
	GtSelReqCD *scd;
	
	obj=obj;
	scd=(GtSelReqCD *)cd;
	tv=(GtTextView *)ud;
	if(scd->target==selatoms[GTDSA_TARGETS])
	{
		scd->target=XA_ATOM;
		scd->fmt=32;
		scd->data=(ubyte *)selatoms;
		scd->len=GTTVA_ATOM_NR;
	}
	else if(scd->target==XA_STRING)
	{
		if(tv->markstr)
		{
			scd->target=XA_STRING;
			scd->fmt=8;
			scd->data=(ubyte *)tv->markstr;
			scd->len=strlen(tv->markstr);
		}
		else
			scd->ok=FALSE;
	}
	return(GTCB_CONT);
}

/************************************************************************\
*
* Lose selection.
*
\************************************************************************/
int GtTextView::LoseMark(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtTextView *tv;
	
	obj=obj;
	cd=cd;
	tv=(GtTextView *)ud;
	if(tv->tflags & GTTF_MARK)
	{
		tv->tflags&=~GTTF_MARK;
		tv->UpdateMark(FALSE);
		tv->Refresh();
	}
	return(GTCB_CONT);
}

/************************************************************************\
*
* Done with selection.
*
\************************************************************************/
int GtTextView::DoneMark(GtBase *obj,GtCDPtr cd,GtUDPtr ud)	
{
	GtTextView *tv;
	
	obj=obj;
	cd=cd;
	tv=(GtTextView *)ud;
	if(tv->markstr)
	{
		free(tv->markstr);
		tv->markstr=NULL;
	}
	return(GTCB_CONT);
}
